#ifndef __CONCEPT_SERVER_H
#define __CONCEPT_SERVER_H

//#define NOLIST

#define MAX_PATHS   255

#ifdef SOURCE_BUILD
    #define LINUX_PATH "/opt/Concept/bin/"
    #define LIBRARY_PATH    "/opt/Concept/bin/concept-core.1.0alpha.so"
    #define INI_PATH    "/opt/Concept/bin/concept.ini"
#endif

#define LOG_DATA(data)                                                                              \
    {                                                                                               \
        time_t now;                                                                                 \
        time(&now);                                                                                 \
        AnsiString sztime=ctime(&now);                                                              \
        char *ptr=sztime.c_str();                                                                   \
        int len=sztime.Length();                                                                    \
        if (ptr[len-1]=='\n')                                                                       \
            ptr[len-1]=0;                                                                           \
        std::ofstream out;                                                                          \
        out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);          \
        out << ptr << ": " << data << "\n";                                                         \
        out.flush();                                                                                \
        out.close();                                                                                \
    }


#include "AnsiString.h"
#include <stdlib.h>

#ifdef _WIN32
    #include <windows.h>
    #include <winsock.h>
    #include <process.h>
#else
    #define SOCKET int
    #define INVALID_SOCKET -1
    #define SOCKET_ERROR -1
    #define DWORD long
    #define LPVOID void*

    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/wait.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <dlfcn.h>   // dlopen
    #include <time.h> // sleep
    #include <pthread.h>
    #include <unistd.h>
    #include <pwd.h>
    #define HANDLE 	void *
    #define HMODULE     HANDLE
#endif
#include "ConceptTypes.h"
#include <stdio.h>
extern "C" {
	#include "semhh.h"
}

typedef int (*GET_VARIABLE_PROC)(void *VDESC, void *CONTEXT, int Depth, char *VariableName, char *buffer, int buf_size);
typedef int (*DEBUGGER_CALLBACK)(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED);

#define PUBLIC_KEY_FILE		"server.certificate"
#define PRIVATE_KEY_FILE	"server.key"
#define PUBLIC_PATH         ".public"
#define PUBLIC_XML          "index.xml"

#define DEFAULT_PORT		    2662
#define DEFAULT_CGI_PORT	    2663
#define MAX_CONNECTIONS		    0x400

#define NO_WEB_ERROR            0
#define ERR_WSA_STARTUP	        100
#define ERR_CREATE_SOCKET       110
#define ERR_BIND_SOCKET         120
#define ERR_LISTEN_SOCKET       130
#define ERR_THREAD_TERMINATE    200

#define MAX_BUFFER				0xFFFF // 64k buffer !
#define MAX_FILENAME			8192
#define MAX_APPLICATIONS        0xFFFF//0x200
#define MAX_PARENTS             0x200
#define DEFAULT_MAX_INTERAPP_BUFFER 1024
#define MAX_IDLE                30 // 30 seconds

#define ERR_INVALID_COMMAND     "Invalid commant"
#define ERR_APPLICATION_ERR     "Application not found, or insufficient rights"
#define ERR_APPLICATION_ERR1    "There are ERRORS. See output dump for details"
#define ERR_APPLICATION_ERR2    "There are RUN-TIME ERRORS. See output dump for details"
#define ERR_APPLICATION_ERRFAILED "Application failed to pass a check point. Infinte loop ?"


typedef struct {
    int pipe_in[2];
    int pipe_out[2];
    unsigned int APID;
    
    AnsiString inter_buf;
    char done;
    HHSEM lock;
#ifdef _WIN32
    LPVOID threadID;
#else
    pthread_t threadID;
#endif
} INTERCOM_ELEMENT;

#define FLAG_DELETED    0
#define FLAG_PARENT     1
#define FLAG_DEBUG      2

typedef struct {
    /*int*/long APID;
    char flag;
    AnsiString key;
} HOST_ELEMENT;

typedef AnsiString (*HostCallBackPath)(char *host);
typedef AnsiString (*ServiceLoader)(int);

class ConceptServer
{
    private:
        static HHSEM sem_lock;

        int PORT;
        int PORT_CGI;


        SOCKET ListenSocket;
        SOCKET CGIListenSocket;
        
        static HostCallBackPath    hcbp;

        INTERCOM_ELEMENT    *APPLICATIONS;
        HOST_ELEMENT        WEBPARENTS[MAX_PARENTS];
        int                 WEBPARENTS_COUNT;

        int INTERCOM_COUNT;
        /*unsigned int*/long NEXT_APID;
        /*unsigned int*/long MINIMUM_AVAILABLE_APID;
        /*unsigned int*/long MAXIMUM_AVAILABLE_APID;
        int INTER_LOCKED;

#ifdef _WIN32
	    static DWORD WINAPI ClientThread(LPVOID);
	    static DWORD WINAPI CGIThread(LPVOID);
        static DWORD WINAPI InterApplicationProcessor(LPVOID server);
	    static DWORD WINAPI RunService_Thread(LPVOID data);
        static DWORD WINAPI ClientAPM(LPVOID P);
#else
	    static LPVOID ClientThread(LPVOID);
	    static LPVOID CGIThread(LPVOID);
        static LPVOID InterApplicationProcessor(LPVOID server);
	    static LPVOID RunService_Thread(LPVOID data);
        static LPVOID ClientAPM(LPVOID P);
#endif

        static AnsiString NormalizePath(AnsiString path);
        static int interapp_buffer;

        int InterMessage(INTERCOM_ELEMENT *IEL, int id, int msg_id, AnsiString message);
        int RouteMessage(int *parameters, char *data, int sender_id);
        int AddAPID(int *pipein, int *pipeout);
        int AddCGIAPID(int pipein, int pipeout);
        int RemoveAPID(int APID, int forced=0);

        static int RunConceptApplication(char *filename, char *arguments, SOCKET sock, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipein, int pipeout, int apid, int parent_apid, char *host_root="./");
        
        static SOCKET CONTACT_DEBUGER(char *hostname, int port);
        static int TALK_TO_CLIENT(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED);
        static int DEBUGGER_TRAP(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED);
        static int RunDebugConceptApplication(char *filename, char *arguments, SOCKET sock, int pipein, int pipeout, int apid, int parent_apid, char *host_root="./");

        static int RunService(char *filename, int pipein, int pipeout, int apid, int parent_apid, char *host_root);
        static int RunService(char *filename);

        static int List(char *path, SOCKET sock);
        static int inter_write(INTERCOM_ELEMENT *IEL, char *buffer, int size);

        struct sockaddr_in sin;
        struct sockaddr_in cgisin;
        static void Print(char *str, int length);

        DWORD threadID;
        HANDLE threadHandle;

        static AnsiString Library;
        static AnsiString Include;
        static AnsiString CertificatePath;
        static AnsiString SafeModePath;

        static void *this_reference;

        static int SAFE_MODE;
    public:
        static AnsiString Service;
        static AnsiString ServicePath;
        static AnsiString LogPath;
        static AnsiString CGITrustedIP;
		static int max_connections;
        static int max_idle;
        static int web_timeout;

        static void SetEnvVar(char *varname, char *value);

        int Host(int APID, char *data, bool debug=false);
        int UnHost(int APID, char *data=0);
        int GetHost(char *data, bool &debug);

        int Init();
        int Start(ServiceLoader sl);
        int StartCGIServer();
        //int Stop();
        ConceptServer(AnsiString Cert, AnsiString Lib, AnsiString Inc, AnsiString Safe, int port=DEFAULT_PORT, int cgiport=DEFAULT_CGI_PORT, HostCallBackPath cb=0, int max_buffer=DEFAULT_MAX_INTERAPP_BUFFER);
        ~ConceptServer(void);
};

#endif //__CONCEPT_SERVER_H
